--[[
-- Copyright 2016 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- See characters/characters.md for a detailed description of a character.
--]]

local M = {}
local scarlettables = require 'scarlettables'
local scarletfunctions = require 'scarletfunctions'
local attack = require 'combat/attack'
local power = require 'combat/powers'
local target = require 'combat/target'

M.unknown = 0
M.male = 1
M.female = 2

local attackpage = target.target(
    function(combatstate, previouspage, ...)
        return attack.attack(
            combatstate,
            previouspage,
            nil,
            scarletfunctions.unpack({...})
        )
    end
)

--[[
-- Returns a character in the game, described in characters/characters.md.
--
-- This function takes a table with all the keys described in characters.md,
-- and returns a deep copy of that table, plus some additional keys needed
-- by the game, but that don't need to be specified by the player:
--  1. statuses - The list of statuses the character is currently afflicted with
--  2. choicecounts - If the character has powers, then this defaults to
--      1 attack and 1 power. Otherwise, it defaults to just 1 attack.
--  3. powers - An empty table if none are provided
--  4. concisename - The name to display for the character when the combat board
--  is in concisemode. Defaults to the first letter of the character's
--  combatname.
-- ]]
function M.character(charactertable)
    local character = scarlettables.copy(charactertable)
    if not character.printedname then
        character.printedname = character.name
    end
    character.printedname = character.printedname or character.name
    character.combatname = character.combatname or character.name
    character.concisename = character.concisename
        or string.sub(character.combatname, 1, 1)
    character.statuses = {}
    if not character.choicecounts then
        character.choicecounts = {
            attack={count=1, page=attackpage},
            power={count=0, page=power.power},
            _keys={"attack", "power"}
        }
        if character.powers then
            character.choicecounts.power.count = 1
        end
    end
    if not character.powers then
        character.powers = {_keys={}}
    end
    character.modifiers = {}
    return character
end

return M
