--[[
-- Copyright 2016 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- This module defines the character creation UI page. The key
-- `M` at the bottom of the module defines the page
-- used to control character creation.
--
-- The body is a GUI for creating a character. The GUI has a field for
-- typing the character's name, followed by a collection of components.
-- Each component represents a different datapoint that defines the
-- character such as class, weapon, or hair color. Each component should
-- have a finite collection of options. The player is then able to scroll
-- through the options with the up and down arrows, or the displayed buttons.
--
-- The choices consist of a back button (triggered by the Escape key) that
-- returns to the previous screen without updating the character, and the done
-- button, which maps to an page that updates the character with the selected
-- value of each component, and then goes to the story display screen.
--]]

local M = {}

local charactercomponents = require 'data/components'
local componentscreen = LOVE and require 'characters/componentscreen'
local fonts = LOVE and require 'ui/fonts'
local log = require 'log'
local love = LOVE and require 'love'
local pages = require 'pages'
local scarletfunctions = require 'scarletfunctions'
local scarlettables = require 'scarlettables'
local uiutils = require 'ui/utilities'
local consoleutils = require 'ui/consoleutils'

local CHARACTER_NAME_ID = "charactername"
local BACK_BUTTON_ID = LOVE and "escape" or "c"
local DONE_BUTTON_ID = "return"

--[[
-- Updates the currently selected index for this choice based on the player's
-- key or button presses.
--
-- selectionState: The current state of the selection, a table with the _keys:
--     value - The index of the currently selected value
--     max - The maximum possible value for `value`
--     min - The minimum possible value for `value`
--     delta - -1 if the player went down through the list,
--      and 1 if the player went up through the list
-- gui: A table representing the state of the game's GUI, should have the same
--  interface as SUIT
-- selectionId: The id of the current widgit
-- activeId: The id of the widgit that is currently activated
-- nextId: The id for the button that moves forward through this component's
--  list of character creation options
-- previousId: The id for the button that moves backwards through this
--  component's list of character creation options
--]]
local function updateSelectionValue(
    selectionState,
    gui,
    selectionId,
    activeId,
    nextId,
    previousId
)
    if activeId ~= selectionId then
        return
    end
    local valueModifier =
        ((gui.getPressedKey() == "up" or gui.isHit(nextId)) and 1)
        or ((gui.getPressedKey() == "down" or gui.isHit(previousId)) and -1)
        or 0
    selectionState.delta = valueModifier
    local newvalue = selectionState.value + valueModifier
    newvalue = newvalue > selectionState.max
        and selectionState.min
        or newvalue < selectionState.min
        and selectionState.max
        or newvalue
    selectionState.value = newvalue
end

--[[
-- Builds a slider for a component of a character.
--
-- componentData: A table containing the data needed to build a character
--  component creation component. The table must have the following keys:
--  id: The id used for the component in the character table (i.e. 'gender'
--      or 'hairstyle')
    label: The label to display for this slider
--  values: The possible values this component can take, and the slider should
--      slide through
--  descriptions: An optional table of additional information to show about each
--      value, keyed on value.
--  shortcut: The key used to select this component
-- uistate: The current UI state
--      In addition to the documented components of the uistate, it is also
--      assumed that the uistate has a (potentially empty) table of tables
--      called `creation`. Each entry of `creation` maps the id of a component
--      to the value currently selected for that component
--      `creation` also has  key `activeId` that tracks the currently selected
--          component
-- labelWidth: The maximum width allowed for this label
--]]
local function buildComponent(componentData, uistate, labelWidth)
    local label = componentData.label
    local gui = uistate.gui
    local shortcut = componentData.shortcut

    local options = uistate.core.gamestate.options
    local labelFont = fonts.button(options.buttonfont)

    local value = uistate.creation[componentData.id]
    local width = labelFont:getWidth(value)
        + fonts.singleCharacterWidth(labelFont)

    local componentsHeight = labelFont:getHeight()
    local componentLabel = string.lower(shortcut) .. ") " .. label
    local labelx, labely = gui.layout:row(labelWidth, componentsHeight)
    gui.Button(
        componentLabel,
        {font = labelFont, align = "left", id=shortcut},
        labelx,
        labely
    )

    gui.Label(
        value,
        {font = labelFont, align="left"},
        gui.layout:col(width, componentsHeight)
    )

    if componentData.descriptions then
        gui.layout:push(labelx, labely + 2.5 * componentsHeight)
        local descriptionfont = fonts.body(options.bodyfont)
        local description = componentData.descriptions[value]
        local width = descriptionfont:getWidth(description)
        gui.Label(
            description,
            {font = descriptionfont, align="left"},
            gui.layout:row(width + 5, componentsHeight)
        )
        gui.layout:pop()
    end
end

--[[
-- Builds the widgits used to display the text input for selecting a character's
-- name.
--
-- The character's name is built using two widgits: a label describing the
-- keyboard shortcut that brings the text input into focus, and the text input
-- that actually takes the name for the player character.
--
-- gui: The SUIT model used for the GUI
-- printedName: A table containing two _keys:
--  text: The text to display in the TextInput box that represents the
--      current state of the player character's name
--  shortcut: The name of the keyboard key that should be used to give the
--      text input widgit focus
--]]
local function buildNameWidgit(uistate, gui, printedName)
    gui.layout:push(love.graphics.getDimensions()/2, 0)

    local titlefont = fonts.title(uistate.core.gamestate.options.titlefont)
    local titlefontoptions = {font=titlefont}
    gui.Label(
        "F1)",
        titlefontoptions,
        gui.layout:col(
            titlefont:getWidth("F1)") + 5,
            titlefont:getHeight()
        )
    )
    local w = titlefont:getWidth(printedName.text) + 5
    local h = titlefont:getHeight()
    local x,y = gui.layout:col(w, h)

    local gainedFocus = gui.getPressedKey() == "f1" or not uistate.creation.activeId
    if gainedFocus then
        uistate.creation.activeId = CHARACTER_NAME_ID
    end
    local selected = uistate.creation.activeId == CHARACTER_NAME_ID
    local boldtitle = fonts.titleBold(
        uistate.core.gamestate.options.titlefont
    )
    local inputoptions = {
        id=CHARACTER_NAME_ID,
        font=selected and boldtitle or titlefont
    }

    gui.Input(
        printedName,
        inputoptions,
        x,
        y
    )
    if gainedFocus then
        gui.setHit(CHARACTER_NAME_ID)
    end
    if selected then
        gui.grabKeyboardFocus(CHARACTER_NAME_ID)
    elseif gui.hasKeyboardFocus(CHARACTER_NAME_ID) then
        gui.keyboardFocus = nil
    end
    gui.layout:pop()
end

-- [[
-- Checks whether or not we should display the specified item when the player
-- is scrolling through clothing to equip.
--
-- We want to make sure that the player can only select a multislot
-- item from the slot group that appears first in the outfit's
-- slots. Otherwise, if the player say selects a shirt first, and
-- then scrolls through the pants/skirts, they might overwrite
-- their chosen shirt with a multislot item.
-- However, we still need the multislot item to appear in the
-- list of values so that *we* can change it when the player selects
-- a mulit-slot item in the appropriate group.
--
-- We also want the *game* to be able to have the player equip whatever
-- it wants, but we don't necessarily want the player to equip that stuff
-- whenever. So we need to make sure that we only display to the player clothing
-- they can actually pick.
--
-- item: The item to check
-- slot: The slot the player is scrolling through
--
--
-- ]]
local function displayable(item, slot)
    return item.slots[1] == slot and item.selectable
end

--[[
-- The same as displayable above, except it returns true if
-- the outfit occupies the specified slot, not just if the 
-- outfit's first slot is the passed in slot
--]]
local function consoledisplayable(item, slot)
    for _, itemslot in ipairs(item.slots) do
        if itemslot == slot then
            return item.selectable
        end
    end
    return false
end

local function getindexofoutfit(selecteditem, clothinglist)
    for itemindex, item in pairs(clothinglist) do
        if item.name == selecteditem.name then
            return itemindex
        end
    end
end

-- [[
-- Updates uistate.clothing with the player's most recent selections, and
-- updates the other selections if the player has (de)selected a multislot
-- article of clothing.
--
-- This function assumes that the for each slot in clothingbyslot, there is 
-- mapping in uistate.creation from the slotname to the name of the outfit
-- selected for that slot
--
-- uistate - The current state of the UI
-- clothingbyslot - A table that maps a slot to the list of clothing that
--  occupies that slot.
-- ]]
local function registerclothing(uistate, clothingbyslot)
    for _, slot in ipairs(clothingbyslot._keys) do
        local selecteditemname = uistate.creation[slot]
        local values = clothingbyslot[slot]
        local selectionindex = 1
        local selecteditem = values[selectionindex]
        for i, item in ipairs(values) do
            if item.name == selecteditemname then
                selectionindex = i
                selecteditem = item
                break
            end 
        end
        if selecteditem.name ~= uistate.clothing[slot].name then
            while not displayable(selecteditem, slot) do
                selectionindex = selectionindex + 1
                if selectionindex > #values then
                    selectionindex = 1
                end
                selecteditem = values[selectionindex]
            end
            for _, selectedslot in ipairs(selecteditem.slots) do
                local olditem = uistate.clothing[selectedslot]
                if olditem then
                    if not uistate.previousitems then
                        uistate.previousitems = {}
                    end
                    local numslots = 0
                    for _, slot in ipairs(olditem.slots) do
                        numslots = numslots + 1
                    end
                    for _, oldslot in ipairs(olditem.slots) do
                        uistate.clothing[oldslot] = nil
                        local previousitem = uistate.previousitems[oldslot]
                        uistate.creation[oldslot] = previousitem
                            and previousitem.item.name
                            or clothingbyslot[oldslot][1].name
                        uistate.clothing[oldslot] = previousitem
                            and previousitem.item
                            or clothingbyslot[oldslot][1]
                        if numslots == 1 then
                            local oldindex = getindexofoutfit(
                                olditem,
                                clothingbyslot[oldslot]
                            )
                            assert(oldindex)
                            uistate.previousitems[oldslot] = {
                                item=olditem,
                                index=oldindex
                            }
                        end
                    end
                end
                uistate.clothing[selectedslot] = selecteditem
                uistate.creation[selectedslot] = selecteditem.name
            end
        end
    end
    for _, slot in ipairs(uistate.clothing._keys) do
        if not uistate.clothing[slot] then
            if uistate.previousitems[slot] then
                local previousitem = uistate.previousitems[slot]
                uistate.clothing[slot] = previousitem.item
                uistate.creation[slot] = previousitem.item.name
            else
                uistate.clothing[slot] = clothingbyslot[slot][1]
            end
        end
    end
end

--[[
-- Groups all the clothing the player can equip by slot
--
-- uistate: The current state of the UI
-- pcclothing: The list of names of clothing the player can wear
--
-- return An ordered table that maps clothing slot to the list of outfits
-- that can be equipped in that slot
--]]
local function groupclothingbyslot(uistate, pcclothing)
    local configuration = uistate.core.gamestate.configuration
    if not uistate.cache then
        uistate.cache = {}
    end
    if not uistate.cache.clothingbyslot then
        local clothingbyslot = {_keys={}}
        local clothing = {}
        for i, name in ipairs(pcclothing) do
            local item = configuration.items[name]
            if not item then assert(false, "Unknown item: ".. name) end
            clothing[i] = item
        end
        for _, item in ipairs(clothing) do
            log.debug(item.name)
            for _, slot in ipairs(item.slots) do
                if not clothingbyslot[slot] then
                    clothingbyslot[slot] = {}
                    table.insert(clothingbyslot._keys, slot)
                    assert(
                        not item.slots[2], -- There is only one slot.
                        "The first item for slot "
                            .. slot
                            .. " name: "
                            .. item.name
                            .. " does not occupy a single slot."
                            .. "\nThe first value for each slot is used as the "
                            .. "default item, and so must occupy one slot to "
                            .. "guarantee valid clothing combinations."
                    )
                end
                table.insert(clothingbyslot[slot], item)
            end
        end
        uistate.cache.clothingbyslot = clothingbyslot
    end
    return uistate.cache.clothingbyslot
end

-- [[
-- Builds the widgits used to let the player select the clothing their character
-- should wear.
--
-- This function adds an ordered table `clothing` to `uistate` that tracks

-- equipped in that slot.
--
-- uistate: The current state of the UI
-- pcclothing: A list of names of clothing
-- slotwidth: The integer width that should be used for each clothing component
-- ]]
local function buildClothingWidgit(uistate, pcclothing, slotwidth)
    local gamestate = uistate.core.gamestate
    local configuration = gamestate.configuration
    local clothingbyslot = groupclothingbyslot(uistate, pcclothing)
    local gui = uistate.gui
    local options = gamestate.options
    local buttonfontsize = fonts.button(options.buttonfont):getHeight()
    gui.layout:push(
        love.graphics.getDimensions() / 8,
        3 * buttonfontsize
    )
    local titlefont = fonts.title(options.titlefont)
    gui.Label(
        "Clothing",
        {font=titlefont},
        gui.layout:row(
            titlefont:getWidth("Clothing") + 5,
            titlefont:getHeight()
        )
    )
    gui.layout:pop()
    local pc = gamestate.characters.pc
    if not uistate.clothing then
        uistate.clothing = {_keys=clothingbyslot._keys}
    end
    for rownum, slot in ipairs(clothingbyslot._keys) do
        local items = clothingbyslot[slot]
        if not uistate.cache.itemnames then
            uistate.cache.itemnames = {}
        end
        if not uistate.cache.itemdescriptions then
            uistate.cache.itemdescriptions = {}
        end
        local itemnames = uistate.cache.itemnames[slot]
        local descriptions = uistate.cache.itemdescriptions[slot]
        local startingvalue = 1
        if not uistate.cache.itemnames[slot] then
            itemnames = {}
            descriptions = {}
            for i, item in pairs(items) do
                if item.name == pc.clothing[slot].name then
                    startingvalue = i
                end
                itemnames[i] = item.name
                descriptions[item.name] = item.description
            end
            if not uistate.creation[slot] then
                uistate.creation[slot] = itemnames[startingvalue]
                uistate.clothing[slot] = pc.clothing[slot]
            end
            uistate.cache.itemnames[slot] = itemnames
            uistate.cache.itemdescriptions[slot] = descriptions
        end
        local componentData = {
            id=slot,
            label=slot,
            values=itemnames,
            descriptions=descriptions,
            shortcut="f" .. (rownum + 1)
        }
        gui.layout:push(
            love.graphics.getDimensions() / 8,
            4 * buttonfontsize * (rownum + 1)
        )
        buildComponent(componentData, uistate, slotwidth)
        gui.layout:pop()
    end
    registerclothing(uistate, clothingbyslot)
end

--[[
-- Builds the button that allows the player to indicate they are done creating
-- their character and would like to move on.
--]]
local function buildDoneButton(uistate)
    local buttonText = "(Enter) Done"
    local buttonfont = fonts.button(buttonfontsize)
    uistate.gui.Button(
        buttonText,
        {id=DONE_BUTTON_ID, font=buttonfont},
        uistate.gui.layout:row(
            fonts.width(buttonfont, buttonText),
            buttonfont:getHeight()
        )
    )
end

--[[
-- Builds the button that allows the player to go back to the previous screen.
--]]
local function buildBackButton(uistate)
    local buttonText = "(Esc) Cancel"
    local buttonfontsize = tonumber(uistate.core.gamestate.options.buttonfont)
    local buttonfont = fonts.button(buttonfontsize)
    uistate.gui.Button(
        buttonText,
        {id=BACK_BUTTON_ID, font=fonts.button(buttonfontsize)},
        uistate.gui.layout:row(
            fonts.width(buttonfont, buttonText),
            buttonfont:getHeight()
        )
    )
end

function M.creationGUI(printedName, uistate, statistics, pcclothing, ...)
    if LOVE then
        return M.loveGUI(printedName, uistate, statistics, pcclothing, ...)
    else
        return M.consoleUI(printedName, uistate, statistics, pcclothing, ...)
    end
end

--[[
-- Builds out the GUI for character creation. The GUI is laid out roughly
-- like this (note that the values displayed are for illustration purposes
-- only):
--                  shortcut) PrintedName
--      Gender: Male !
--      Manipulator: Phantom !
--      Description: A gravity-manipulator dangerous at close quarters
--      Weapon: Knife
--      Description: A small blade dangerous at close quarters, but not when
--      further away
--      Strength: 2 Speed: 2 Willpower: 2 Power: 30/30
--
-- Where ! represents a scrollbar, and _ indicates that that field is
-- takes text input.
--
-- printedName: A table with a two keys:
--      text: The text to populate the Input widgit with
--      shortcut: A key that when pressed allows the player to provide input
--      to the widgit
-- uistate: The current state of the UI, this will be augmented by the
--  function. However, none of the uistate's canonical mappings will be modified.
-- statistics: An ordered table of strings representing the character's
--  statistics
-- pcclothing: A list of the names of the clothing the PC can wear
-- ... : All subsequent arguments are character components. They should be
--  tables that describe a character choice (i.e. class or gender). The tables
--  must include:
--      id: The id to use for the associated widgit
--      label: The text to use to label the choice
--      values: The list of possible values for that choice
--      shortcut: A keyboard shortcut that when pressed, allows the
--      player to manipulate the choice's slider with the arrow _keys
--
-- In order to track the choices made by the player for each component, the
-- uistate is augmented with a mapping from component.id to the value currently
-- selected by the player for that component. This table is called `creation`.
--
-- Returns a list containing the metadata of each widgit constructed,
-- starting with the widgits used to build the text input, then the data for
-- each widgit for each statistic, and finally the widgit for the statistic
--
-- Note that the shortcut _keys are not allowed to be printable characters.
-- Otherwise, if a user types a key while the name widgit is in focus, we
-- don't know if they want that character in the player's name, or to change
-- focus! It is recommended that the function _keys are used as keyboard
-- shortcuts.
--]]
function M.loveGUI(printedName, uistate, statistics, pcclothing, ...)
    local gui = uistate.gui
    if not uistate.creation then
        uistate.creation = {}
    end
    -- Used to remember things that don't change across frames, like
    -- the clothing player can pick from
    if not uistate.cache then
        uistate.cache = {}
    end
    local componentList = {...}
    local options = uistate.core.gamestate.options
    local buttonboldfont = fonts.buttonBold(options.buttonfont)
    local maxNameWidth = 0
    for _, component in ipairs(componentList) do
        local label = component.shortcut .. ") " .. component.label
        local labelwidth = fonts.width(buttonboldfont, label)
        if labelwidth > maxNameWidth then
            maxNameWidth = labelwidth
        end
    end

    buildNameWidgit(uistate, gui, printedName)

    buildClothingWidgit(uistate, pcclothing, maxNameWidth)

    local buttonfont = fonts.button(options.buttonfont)
    gui.layout:push(
        love.graphics.getDimensions() / 8,
        3 * buttonfont:getHeight() * (#uistate.clothing._keys + 4)
    )
    local titlefont = fonts.title(options.titlefont)
    gui.Label(
        "Commands",
        {font=titlefont},
        gui.layout:row(
            titlefont:getWidth("Commands") + 5,
            titlefont:getHeight()
        )
    )
    buildBackButton(uistate)
    buildDoneButton(uistate)
    gui.layout:pop()

    gui.layout:push(
        love.graphics.getDimensions() / 2,
        3 * buttonfont:getHeight()
    )
    gui.Label(
        "Personalizations",
        {font=titlefont},
        gui.layout:row(
            titlefont:getWidth("Personalizations") + 5,
            titlefont:getHeight()
        )
    )
    gui.layout:pop()
    local pc = uistate.core.gamestate.characters.pc
    for rowNum, componentData in ipairs(componentList) do
        local componentid = componentData.id
        if not uistate.creation[componentid] then
            for i, value in ipairs(componentData.values) do
                if value == pc.attributes[componentid] then
                    uistate.creation[componentid] = value
                end
            end
            assert(
                uistate.creation[componentid],
                "Unknown component value: "
                    .. (pc[componentid] or "nil")
                    .. " for component: " .. componentid .. "."
                    .. " Check the pc's configuration in data/characters.lua."
                    .. " Creation: " .. scarlettables.string(uistate.creation)
            )
        end
        gui.layout:push(
            love.graphics.getDimensions() / 2,
            3 * buttonfont:getHeight() * (rowNum + 1)
        )
        buildComponent(componentData, uistate, maxNameWidth)
        gui.layout:pop()
    end
end

--[[
-- Given the currently available clothing, and personalizations, builds
-- a table that maps numbers (which the player can input) to tables
-- containing:
--  slot/id: The clothing slot/personalization id we're modifying
--  values: Either a list of clothing items that can be selected for this
--  slot, or a list of strings containing options for the given personalization
--
-- clothingbyslot: An ordered table that groups clothing by the slots they can
--  occupy
-- personalizations: A list of personalization options available to the player
--
-- returns A table that maps numbers to the choice that number selects
--]]
local function buildchoices(clothingbyslot, personalizations)
    local keycount = 2
    local choices = {}
    for _, slot in ipairs(clothingbyslot._keys) do
        choices[keycount] = {slot=slot, values=clothingbyslot[slot]}
        keycount = keycount + 1
    end
    for _, personalization in ipairs(personalizations) do
        choices[keycount] = {
            id=personalization.id,
            values=personalization.values
        }
        keycount = keycount + 1
    end
    return choices
end


--[[
-- Builds the clothing and personalization lines to display on the
-- character creation screen.
--
-- chosenclothing: The clothing currently being worn by the player
-- chosenpersonalizations: The player's current personalizations
-- choices: A table that maps slots to all possible articles of clothing that
--  can occupy that slot, and that maps personalization ids to all possible values for
--  that personalizatio.
--
-- return A list of lines to display the player's clothing, and a list of lines to
--  display the player's personalizations
--]]
local function buildlines(
    chosenclothing,
    chosenpersonalizations,
    choices
)
    local clothinglines = {}
    local personalizationlines = {}
    for keycount=2,#choices do
        local choice = choices[keycount]
        if choice.slot then
            local slot = choice.slot
            local outfit = chosenclothing[slot]
            table.insert(
                clothinglines,
                string.format("(%d) %s: %s", keycount, slot, outfit.name)
            )
        else
            local id = choice.id
            local personalization = chosenpersonalizations[id]
            table.insert(
                personalizationlines,
                string.format("(%d) %s: %s", keycount, id, personalization)
            )
        end
    end
    return clothinglines, personalizationlines
end

--[[
-- Given a list of lines describing clothing to display, and a list
-- of lines describing the personalizations to display, combines them
-- together into a single character creation screen.
--
-- clothinglines: A list of strings describing the clothing to display
-- personalizationlines: A list of strings describing the personalizations to
--  display
--
-- returns A list of lines to display in the character creation screen
--]]
local function buildcreationlines(clothinglines, personallines, maxclothingline)
    local lines = {}
    local numlines = math.max(#clothinglines, #personallines)
    for i=1,numlines do
        local clothingline = clothinglines[i] or ""
        clothingline = clothingline 
            .. string.rep(" ", maxclothingline - #clothingline)
        local personalizationline = personallines[i] or ""
        lines[i] = table.concat({clothingline, personalizationline}, "\t")
    end
    return lines
end

--[[
-- Builds the character creation screen to display.
--
-- clothinglines: The list of lines to display for clothing
-- personallines: The list of lines to display for personalizations
-- charactername: The current name of the character
--]]
local function buildcreationscreen(clothinglines, personallines, charactername)
    local maxclothingline = 0
    for _, line in ipairs(clothinglines) do
        local linelength = #line
        if maxclothingline < linelength then
            maxclothingline = linelength
        end
    end
    local lines = buildcreationlines(
        clothinglines,
        personallines,
        maxclothingline
    )
    local title = "------------ Character Screen --------------"
    local name = "(1) Name: " .. charactername
    local clothingheading = "Clothing:"
    clothingheading = clothingheading
        .. string.rep(" ", math.max(0, maxclothingline - #clothingheading))
    local headings = {clothingheading, "Personalizations:"}
    local commands = table.concat({"(C)ancel", "(Enter) Done"}, " ")
    return table.concat(
        {
            title,
            name,
            "",
            table.concat(headings, "\t"),
            table.concat(lines, "\n"),
            "",
            commands
        },
        "\n"
    )
end

--[[
-- Equips the specified outfit.
--
-- outfit: The outfit to equip
-- chosenclothing: A table mapping slots to the clothing currently equipped
-- clothingbyslot: A table mapping slots to the list of clothing that can
--  be equipped in this slot
--]]
local function equipoutfit(outfit, chosenclothing, clothingbyslot)
    for _, slot in ipairs(outfit.slots) do
        local oldoutfit = chosenclothing[slot]
        for _, oldslot in ipairs(oldoutfit.slots) do
            chosenclothing[oldslot] = clothingbyslot[oldslot][1]
        end
        chosenclothing[slot] = outfit
    end
end

--[[
-- Displays a list of clothing options to the user for this slot, and updates the
-- appropriate slots based on the player's choice.
--
-- chosenclothing: The table mapping slots to the currently worn clothing
-- choice: A table with two keys:
--  slot: The slot currently being modified
--  command: (Optional) The string used to filter the results
--  values: A list of clothing that can occupy this slot
-- clothingbyslot: A table mapping slots to the list of clothing that can occupy that slot
--
-- return true if chosenclothing was modified, false otherwise
--]]
local function updateclothing(chosenclothing, choice, clothingbyslot)
    local visibleclothing = {}
    local clothingbyname = {}
    for _, outfit in ipairs(choice.values) do
        if consoledisplayable(outfit, choice.slot) then
            table.insert(visibleclothing, outfit)
            clothingbyname[outfit.name] = outfit
        end
    end
    local visiblevalues = {}
    for _, value in ipairs(choice.values) do
        if consoledisplayable(value, choice.slot) then
            table.insert(visiblevalues, value)
        end
    end
    if #visiblevalues == 1 then
        equipoutfit(visiblevalues[1], chosenclothing, clothingbyslot)
        return true
    end
    local valuenames = {}
    for i, outfit in ipairs(visibleclothing) do
        valuenames[i] = string.format(
            "(%d) %s: %s",
             i,
             outfit.name,
             outfit.description
        )
    end
    local header = string.format(
        "------------ %s --------------",
        choice.search or choice.slot
    )
    local commands = "(Enter) Do nothing."
    local values = table.concat(valuenames, "\n")
    local direction = "Number, name or search string:"
    print(table.concat(
        {
            header,
            "",
            values,
            "",
            commands,
            direction
        },
        "\n"
    ))
    while true do
        local command = io.stdin:read()
        -- If command is empty, that means the
        -- player pressed return.
        if #command == 0 then
            return false
        end
        local commandnumber = tonumber(command)
        local outfit
        if commandnumber then
           outfit = visibleclothing[commandnumber]
        else
            outfit = clothingbyname[command]
            if not outfit then
                return consoleutils.patternmatch(
                    choice,
                    command,
                    function(value) return value.name end,
                    function(achoice)
                       return updateclothing(
                            chosenclothing, 
                            achoice, 
                            clothingbyslot
                        )
                    end
                )
            end
        end
        if outfit then
            equipoutfit(outfit, chosenclothing, clothingbyslot)
            return true
        end
    end
end

--[[
-- Performs pattern matching of the given command against the names of
-- the available personalization option, and lists all options that match the 
-- pattern.
--
-- The pattern is defined as ".*command.*" with `command` properly
-- escaped.
--
-- chosenpersonalizations: The table mapping slots to the currently worn clothing
-- choice: A table with two keys:
--  id: The slot currently being modified
--  command: (Optional) The string used to filter the results
--  values: A list of clothing that can occupy this slot
-- command: The search string provided by the user
--
-- return true if chosenclothing was modified, false otherwise
--]] 
local function patternmatchpersonalizations(chosenpersonalizations, choice, command)
    local escaped = string.gsub(command, "%-", "%%-")
    local pattern = ".*" .. escaped  .. ".*"
    local matchingoptions = {}
    for _, option in ipairs(choice.values) do
        if string.match(option, pattern) then
            table.insert(matchingoptions, option)
        end
    end
    local updated = updatepersonalization(
        chosenpersonalizations,
        {id=choice.id, search=command, values=matchingoptions}
    )
    if not updated then
        return updatepersonalization(chosenpersonalizations, choice)
    end
    return updated
end

--[[
-- Displays a list of personalization options to the user for this personalization
-- and updates it based on the player's choice.
--
-- chosenpersonalizations: The settings the player has currently chosen for each
--  personalization
-- choice: A table with two keys:
--  id: The personalization id currently being modified
--  values: A list of personalization values (i.e. strings) for this choice
--]]
local function updatepersonalization(chosenpersonalizations, choice)
    if #choice.values == 1 then
        chosenpersonalizations[choice.id] = choice.values[1]
        return true
    end
    local valuenames = {}
    for i, value in ipairs(choice.values) do
        valuenames[i] = string.format("(%d) %s", i, value)
    end
    local header = string.format("------------ %s --------------", choice.id)
    local commands = "(Enter) Do nothing.)"
    local direction = "Number, name or search string:"
    print(table.concat(
        {
            header,
            table.concat(valuenames, "\n"),
            "",
            commands
        }, 
        "\n"
    ))
    while true do
        local command = io.stdin:read()    
        if #command == 0 then
            return false
        end
        local commandnumber = tonumber(command)
        local value
        if commandnumber then
            value = choice.values[commandnumber]
        else
            for _, personalization in ipairs(choice.values) do
                if command == personalization then
                    value = personalization
                end
            end
            if not value then
                return consoleutils.patternmatch(
                    choice,
                    command,
                    function(value) return value end,
                    function(achoice)
                        return updatepersonalization(
                            chosenpersonalizations, 
                            achoice
                        )
                    end
                )
            end
        end
        if value then
            chosenpersonalizations[choice.id] = value
            return true
        end
    end
end

--[[
-- When the user inputs a number, asks the user to select a value for the appropriate
-- slot, and then updates either chosenclothing or chosenpersonalizations as appropriate.
--
-- uistate: The current state of the UI
-- printedname: A table containing at least the key `text` which is the current name of the
--  player
-- chosenclothing: A table mapping slots to the clothing the player is wearing
-- chosenpersonalizations: A table mapping personalization ids to the value for that
--  personalization
-- choices: A table that maps slots to all possible articles of clothing that
--  can occupy that slot, and that maps personalization ids to all possible values for
--  that personalizatio.
-- clothingbyslot: A table mapping clothing slots to the list of clothing equippable in
--  that slot
--]]
local function processinput(
    uistate,
    printedname, 
    chosenclothing, 
    chosenpersonalizations, 
    choices,
    clothingbyslot
)
    if uistate.command then
        local commandnumber = tonumber(uistate.command)
        if commandnumber and commandnumber > 1 then
            local choice = choices[commandnumber]
            if choice and choice.slot then
                updateclothing(chosenclothing, choice, clothingbyslot)
            elseif choice then
                local id = choice.id
                return updatepersonalization(chosenpersonalizations, choice)
            end
        elseif commandnumber == 1 then
            print("------------ Provide a Name --------------")
            local name = io.stdin:read()
            if #name > 0 then
                printedname.text = name
                return true
            end
        end
    end
end

--[[
-- Builds a UI for character creation in the console. The UI is set up as follows:
--                  (1) Name: "Scarlette"
--
--  Clothing:                           Personalizations:
--  (2) Upper: Shirt                    (5) Gender: Male
--  (3) Lower: Jeans                    (6) Height: Tall
--  (4) Underwear: Boxers               ...
--
--
-- Each choice transitions to a different UI, where we list the options, and the player
-- selects one.
--
-- printedName: A table with a two keys:
--      text: The text to populate the Input widgit with
--      shortcut: A key that when pressed allows the player to provide input
--      to the widgit
-- uistate: The current state of the UI, this will be augmented by the
--  function. However, none of the uistate's canonical mappings will be modified.
-- statistics: An ordered table of strings representing the character's
--  statistics
-- pcclothing: A list of the names of the clothing the PC can wear
-- ... : All subsequent arguments are character components. They should be
--  tables that describe a character choice (i.e. class or gender). The tables
--  must include:
--      id: The id to use for the associated widgit
--      label: The text to use to label the choice
--      values: The list of possible values for that choice
--      descriptions: An optional table of additional information to show about
--          each value, keyed on value.
--      shortcut: A keyboard shortcut that when pressed, allows the
--          player to manipulate the choice's slider with the arrow _keys
--]]
function M.consoleUI(printedName, uistate, statistics, pcclothing, ...)
    local clothingbyslot = groupclothingbyslot(uistate, pcclothing)
    local pc = uistate.core.gamestate.characters.pc
    if not uistate.clothing then
        uistate.clothing = {_keys=clothingbyslot._keys}
        for _, slot in ipairs(clothingbyslot._keys) do
            uistate.clothing[slot] = pc.clothing[slot]
        end
    end
    local chosenclothing = uistate.clothing
    local personalizations = {...}
    if not uistate.personalizations then
        uistate.personalizations = {}
        for _, personalization in ipairs(personalizations) do
            local id = personalization.id
            uistate.personalizations[id] = pc.attributes[id]
        end
    end
    local chosenpersonalizations = uistate.personalizations

    local choices = buildchoices(clothingbyslot, personalizations)
    processinput(
        uistate,
        printedName,
        chosenclothing,
        chosenpersonalizations,
        choices,
        clothingbyslot
    )

    local clothinglines, personalizationlines = buildlines(
        chosenclothing,
        chosenpersonalizations,
        choices
    )
    local creationscreen = buildcreationscreen(
        clothinglines,
        personalizationlines,
        printedName.text
    )

    print(creationscreen)
    uistate.creation = uistate.personalizations
end

--[[
-- Builds a GUI for character creation set up as follows:
--                      "Scarlette"
--      Clothing:               Personalizations:
--      Shirt: T-Shirt /\ \/    Gender: Male /\ \/
--
--      Statistics:
--      Strength: 2
--      Speed: 2
--      Willpower: 2
--      Energy: 30/30
--]]
function M.buildcreationgui(uistate)
    local gamestate = uistate.core.gamestate
    if not uistate.characterName then
        uistate.characterName = {
            text = gamestate.characters.pc.printedname
        }
    end

    local configuration = gamestate.configuration

    local gui = M.creationGUI(
        uistate.characterName,
        uistate,
        gamestate.characters.pc.statistics,
        charactercomponents.clothing(configuration.items),
        scarletfunctions.unpack(charactercomponents.buildcomponents())
    )
end

--[[
-- Removes the temporary table used to track character creation data.
-- This should be invoked in every transformer in every page that transitions
-- from this page.
--
-- uistate: The state whose creation table needs to be dropped
--]]
local function clearcreation(uistate)
    uistate.characterName = nil
    uistate.creation = nil
    uistate.cache = nil
    uistate.clothing = nil
    uistate.personalizations = nil
    uistate.previousitems = nil
end

--[[
-- Returns a UI story page that updates the state with the first page for the
-- game ("start"), and a transformer that applies the player's choices to the
-- character being created.
--
-- Note: This function also clears the `creation` key since that key is only
-- used by functions defined in this module. Once we transition from character
-- creation, `creation` is no longer needed.
--]]
local function done(nextpage)
    local components = charactercomponents.buildcomponents()
    local function applySelections(uistate)
        local pc = uistate.core.gamestate.characters.pc
        pc.clothing = uistate.clothing
        for _, component in pairs(components) do
            pc.attributes[component.id] = uistate.creation[component.id]
        end
        local printedname = uistate.characterName.text
        pc.printedname = printedname
        if not pc.combatname or pc.combatname == pc.name then
            pc.combatname = printedname
        end
        clearcreation(uistate)
    end

    local function applyConsoleSelections(uistate)
        local pc = uistate.core.gamestate.characters.pc
        pc.clothing = uistate.clothing
        for id, option in pairs(uistate.personalizations) do
            pc.attributes[id] = option
        end
        pc.printedname = uistate.characterName.text
        if not pc.combatname or pc.combatname == pc.name then
            pc.combatname = pc.printedname
        end
        clearcreation(uistate)
    end

    return pages.page(
        nil,
        function() return nextpage end,
        nil,
        LOVE and applySelections or applyConsoleSelections
    )
end

--[[
-- Builds the page choices for the creation screen.
--
-- uistate: The current state of the UI
-- previouspage: The page to go back to when canceling
-- nextpage: The page to go to after applying the changes
--]]
local function creationchoices(uistate, previouspage, nextpage)
    local charactercreationscreen = M.charactercreationscreen(
        nextpage,
        previouspage
    )
    local increasefont = LOVE
        and fonts.increasefontsizepage(charactercreationscreen)
        or nil
    local increasefontplus = LOVE
        and fonts.increasefontsizepage(charactercreationscreen, nil, true)
        or nil
    local decreasefont = LOVE
        and fonts.decreasefontsizepage(charactercreationscreen)
        or nil

    local clearcreationpage = pages.page(
        nil,
        function() return previouspage end,
        nil,
        clearcreation
    )
    local commands = {
        [BACK_BUTTON_ID] = clearcreationpage,
        [DONE_BUTTON_ID] = done(nextpage),
        ["="] = increasefont,
        ["+"] = increasefontplus,
        ["-"] = decreasefont,
        _keys = {BACK_BUTTON_ID, DONE_BUTTON_ID}
    }
    if LOVE then
        table.insert(commands._keys, "=")
        table.insert(commands._keys,"+")
        table.insert(commands._keys, "-")
        local components = charactercomponents.buildcomponents()
        for _, component in ipairs(components) do
            commands[component.shortcut] = componentscreen.page(
                component, 
                charactercreationscreen
            )
            table.insert(commands._keys, component.shortcut)
        end
        local clothing = charactercomponents.clothing()
        local clothingbyslot = groupclothingbyslot(uistate, clothing)
        local fcount = 2
        for _, slot in ipairs(clothingbyslot._keys) do
            local playeroutfits = {}
            for _, outfit in ipairs(clothingbyslot[slot]) do
                if displayable(outfit, slot) then
                    table.insert(playeroutfits, outfit)
                end
            end
            local component = {
                id=slot,
                label=slot,
                values=playeroutfits
            }
            local key = "f" .. tostring(fcount)
            fcount = fcount + 1
            commands[key] = componentscreen.page(
                component, 
                charactercreationscreen
            )
            table.insert(commands._keys, key)
        end
    end
    return commands
end

-- [[
-- Builds a screen for checking a character's status, and updating things like
-- their outfit, their appearance and their name.
--
-- nextpage: The page to transition to after we're done with the character
-- creation screen, defaults to the options page
-- previouspage: The page to transition to after canceling, defaults to nextpage
--
-- returns A UI page for character creation
-- ]]
function M.charactercreationscreen(nextpage, previouspage)
    previouspage = previouspage or nextpage
    local function choices(uistate) 
        return creationchoices(uistate, previouspage, nextpage)
    end
    local creationui = pages.page(nil, nil, choices)
    creationui.drawgui = M.buildcreationgui
    return creationui
end

return M
