--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- This module defines the target action used by Scarlet. The key function
-- here is `target` which allows us to select one or more targets for any
-- action that targets combat participants.
--]]
--
local M = {}

local combatoperations = require 'combat/operations'
local F = require 'f-strings/F'
local pages = require 'pages'
local scarletfunctions = require 'scarletfunctions'

local function executeaction(combatstate, targets, previouspage, action)
    local actionpage = action(
        combatstate,
        previouspage,
        scarletfunctions.unpack(targets)
    )
    return pages.page(
        nil,
        function() return actionpage end,
        nil,
        function(combatstate1)
            combatstate1.selections = {}
        end
    )
end

--[[
-- Handles the user selecting one or more targets for an action.
--
-- action - A function that takes the combatstate, the page to transition
--  to if the selected target is invalid, and a variable number of target keys,
--  and returns the action to execute
-- numtargets - How many targets this action needs, defaults to 1
--
-- return The next action to execute
--]]

function M.target(action, numtargets)
    numtargets = numtargets or 1
    local function choices(combatstate)
        local targets = combatstate.selections
        local remainingtargets = numtargets - #targets
        local selectstring = F"(#) Select {remainingtargets} targets."
        assert(selectstring)
        local previouspage = M.target(action, numtargets)
        local targetingchoices = {_keys={}}
        if remainingtargets > 0 then
            targetingchoices[selectstring]=M.target(action, numtargets)
            table.insert(targetingchoices._keys, selectstring)
        end
        if remainingtargets < numtargets then
            targetingchoices["(return)confirm"] = executeaction(
                combatstate,
                targets,
                previouspage,
                action
            )
            table.insert(targetingchoices._keys, "(return)confirm")
        end
        return targetingchoices
    end
    local function automatics(combatstate)
        local targets = combatstate.selections
        local remainingtargets = numtargets - #targets
        local previouspage = M.target(action, numtargets)
        if remainingtargets == 0 then
            return executeaction(combatstate, targets, previouspage, action)
        end
    end
    return pages.page(nil, automatics, choices)
end

function M.buildcharacterlabels(combatstate, characters, keyprefix)
    keyprefix = keyprefix or ""
    local labels = {}
    -- (key)name:statuses:energy/maxenergy
    local labeltemplate = "(%s)%s:%s:%d/%d"
    for i, character in ipairs(characters) do
        local key = keyprefix .. i
        labels[character.name] = string.format(
            labeltemplate,
            M.selected(combatstate.selections, key) and "X" or key,
            character.combatname,
            combatoperations.buildstatuslabels(character.statuses),
            character.statistics.energy,
            character.statistics.maxenergy
        )
    end
    return labels
end

function M.selected(selections, key)
    for _, selection in pairs(selections) do
        if selection == key then
            return true
        end
    end
    return false
end

return M
