--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- Implements Poseidon's powers. Poseidon manipulates water, and has two
-- modes: wet and frozen. When wet, he uses his water based powers, which are
-- a balance of offensive and defensive nature. When frozen, he uses his ice
-- powers, which are mostly about making him a nigh invulnerable bruiser.
--]]

local M = {}

local attacks = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local log = require 'log'
local pages = require 'pages'
local powers = require 'combat/powers'
local powerutils = require 'data/powerutils'
local statuses = require 'data/statuses'
local statusops = require 'combat/statuses'

---------------------------- Hydrohammer ---------------------
--[[
-- Poseidon hits an enemy with a ram of water, doing
-- power based damage, and inflicting Wet
-- on the target.
--]]
local function hydrohammerdamage(user, target, combatstate)
    local hammermsg = user.combatname .. " swirls water into a battering "
        .. " ram and smashes it "
        .. "into " .. target.combatname .. "'s" .. " chest."
    local function formula(attacker, defender)
        local attack = attacks.attackstat(attacker) 
        if statusops.statusduration(defender, statuses.wet) then
            attack = 2 * attack
        end
        defense = attacks.defensestat(defender)
        return math.max(0, attack - defense)
    end
    local function hydrohammerdamage(attacker, defender, combatstate1)
        local damage, retaliation, description = attacks.defaultdamage(
            attacker,
            defender,
            combatstate1,
            formula
        )
        description[1] = hammermsg .. " " .. description[1]
        return damage, retaliation, description
    end
    attacks.damage(combatstate, hydrohammerdamage, {target}, user)
end

local function hydrohammer(combatstate)
    local user = combatstateops.activecharacter(combatstate)
    local activeopponents = combatstateops.activeopponents(combatstate)
    local target = powerutils.median(activeopponents, combatstate.round)
    if not target then
        return
    end
    hydrohammerdamage(user, target, combatstate)
    local duration = attacks.statistic(user, "power")
    statusops.addstatus(target, statuses.wet, duration, user, nil, combatstate)
end

M.hydrohammer = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        hydrohammer(combatstate)
        powerutils.inflictcost(combatstate, M.hydrohammer)
    end
)

function M.hydrohammer.aiexecution()
    return M.hydrohammer
end

M.hydrohammer.damage = 1
M.hydrohammer.targets = combatstateops.activeopponents
M.hydrohammer.cost = 6
M.hydrohammer.numtargets = 1
M.hydrohammer.stats = {"power", "speed"}
M.hydrohammer.description = "Poseidon does power - (target.speed - speed) "
    .. "damage to one target and inflicts wet. Double damage if the target "
    .. "already wet."
M.hydrohammer.aiexecution = function() return M.hydrohammer end

---------------------------- Hydroslide ---------------------

--[[
-- Poseidon uses a slide of water to move around rapidly, increasing his
-- defense by power * .5, and granting him an extra attack.
--]]
M.hydroslide = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local duration = attacks.statistic(character, "power")
        local msg = character.combatname 
            .. " rockets into the air on a slide of water!"
        table.insert(combatstate.combatlog, msg)
        statusops.addstatus(character, statuses.hydroslide, duration, character, nil, combatstate)
        statusops.addstatus(character, statuses.wet, duration, character, nil, combatstate)
        powerutils.inflictcost(combatstate, M.hydroslide)
    end
)

M.hydroslide.damage = powers.NEITHER
M.hydroslide.targets = combatstateops.activeallies
M.hydroslide.cost = 2
M.hydroslide.numtargets = 0
M.hydroslide.stats = {"power"}
M.hydroslide.description = "Poseidon increases his defense by power / 2, and "
    .. "his Attack count by 1."
M.hydroslide.aiexecution = function() return M.hydroslide end

-------------------- Chionfist ---------------------------------
local function chionfist(combatstate)
    local character = combatstateops.activecharacter(combatstate)
    local opponents = combatstateops.activeopponents(combatstate)
    local target = powerutils.median(opponents, combatstate.round)
    if not target then 
        return 
    end
    local function formula(attacker, defender)
        local attack = 2 * attacks.attackstat(attacker)
        local defense = attacks.defensestat(defender)
        return math.max(2, attack - defense)
    end
    local chionmsg = character.combatname 
        .. " punches " 
        .. target.combatname
        .. " with an icy fist!"
    local function chiondamage(attacker, defender, combatstate)
        local damage, retaliation, description = attacks.defaultdamage(
            attacker,
            defender,
            combatstate,
            formula
        )
        description[1] = chionmsg .. " " .. description[1]
        return damage, retaliation, description
    end
    attacks.damage(combatstate, chiondamage, {target}, character)
end

--[[
-- Poseidon punches an enemy with an icy fist, doing 2 * attack damage.
--]]
M.chionfist = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        chionfist(combatstate)
        powerutils.inflictcost(combatstate, M.chionfist)
    end
)

M.chionfist.damage = 1
M.chionfist.targets = combatstateops.activeopponents
M.chionfist.numtargets = 1
M.chionfist.aiexecution = function() return M.chionfist end
M.chionfist.cost = 3
M.chionfist.stats = {"strength", "speed"}
M.chionfist.description = "Poseidon punches an enemy with an icy fist, "
    .. "attacking as if his attack stat were twice as large. May only"
    .. " be used when Frozen."
M.chionfist.aiexecution = function() return M.chionfist end

-------------------- Chionskin -----------------------------------------
--[[
-- Poseidon can give himself power * 5 retaliation.
--]]
M.chionskin = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local power = attacks.statistic(character, "power")
        character.secondarystatistics.retaliation.chionskin = power * 5
        local msg = "The air around " .. character.combatname .. " frosts."
        table.insert(combatstate.combatlog, msg)
        powerutils.inflictcost(combatstate, M.chionskin)
    end
)

M.chionskin.damage = 0
M.chionskin.targets = combatstateops.activeallies
M.chionskin.numtargets = 0
M.chionskin.aiexecution = function() return M.chionskin end
M.chionskin.cost = 4
M.chionskin.stats = {"power"}
M.chionskin.description = " Poseidon "
    .."gains retaliation equal to 5 * power. May only be used while Frozen."
M.chionskin.aiexecution = function() return M.chionskin end

-- Hydrogater is a summoning spell that allows Poseidon to use half his current
-- energy to effectively create a copy of himself.
M.hydrogator = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local msg = character.combatname 
            .. " swirls a ball of water into a gator!"
        table.insert(combatstate.combatlog, msg)
        powerutils.summon(combatstate, character, "Hydrogator", "H")
        powers.applycost(character, M.hydrogator)
    end
)

M.hydrogator.damage = 0
M.hydrogator.targets = combatstateops.activeallies
M.hydrogator.cost = 15
M.hydrogator.numtargets = 0
M.hydrogator.stats = {"strength", "speed", "power"}
M.hydrogator.description = "Poseidon weaves his water into an autonomous watery alligator with all "
    .. " his strength and powers. However, he does have to split his energy between himself and "
    .. " his gator. Costs *15* energy."
M.hydrogator.aiexecution = function() return M.hydrogator end

return M
