--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- Implements Tempestas' powers. Tempestas focuses primarily on damage control
-- (though not necessarily reduction). She can increase ally damage, redirect
-- damage, decrease enemy damage, etc.
--]]

local attacks = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local combatoperations = require 'combat/operations'
local F = require 'f-strings/F'
local log = require 'log'
local pages = require 'pages'
local powers = require 'combat/powers'
local powerutils = require 'data/powerutils'
local scarlettables = require 'scarlettables'
local statuses = require 'data/statuses'
local statusops = require 'combat/statuses'

local M = {}

local function windinflict(user, target, combatstate, status, description)
    local duration = math.floor(attacks.statistic(user, "power") / 3)
    statusops.addstatus(target, status, duration, user, combatstate)
end

M.headwind = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local user = combatstateops.activecharacter(combatstate)
        local duration = math.floor(attacks.statistic(user, "power") / 3)
        local opponents = combatstateops.opponents(combatstate, user)
        local usercombatname = user.combatname
        for _, opponent in ipairs(opponents) do
            local targetcombatname = opponent.combatname
            local msg =  usercombatname .. " buffets " .. targetcombatname
                .. " with gale-force winds."
            table.insert(combatstate.combatlog, msg)
            statusops.addstatus(opponent, statuses.headwind, duration, user, combatstate)
        end
        powerutils.inflictcost(combatstate, M.headwind)
    end
)

M.headwind.cost = 3
M.headwind.damage = 0
M.headwind.targets = combatstateops.activeopponents
M.headwind.numtargets = 0
M.headwind.stats = {"power"}
function M.headwind.description(character)
    local duration = math.floor(attacks.statistic(character, "power") / 3)
    return F"Tempestas reduces the attack of all enemies by {duration} for "
        .. F"{duration} turns."
end

M.tailwind = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local user = combatstateops.activecharacter(combatstate)
        local duration = math.floor(attacks.statistic(user, "power") / 3)
        local allies = combatstateops.allies(combatstate, user)
        local usercombatname = user.combatname
        for _, ally in ipairs(allies) do
            statusops.addstatus(
                ally,
                statuses.tailwind,
                duration,
                user,
                combatstate
            )
            local targetcombatname = ally.combatname
            local msg =  usercombatname .. " coils gale-force winds behind "
                .. targetcombatname .. "."
            table.insert(combatstate.combatlog, msg)
        end
        powerutils.inflictcost(combatstate, M.headwind)
    end
)

M.tailwind.cost = 3
M.tailwind.damage = 0
M.tailwind.targets = combatstateops.activeallies
M.tailwind.numtargets = 0
M.tailwind.stats = {"power"}
function M.tailwind.description(character)
    local duration = math.floor(attacks.statistic(character, "power") / 3)
    return F"Tempestas increases the attack of all allies by {duration} for "
        .. F"{duration} turns."
end

local function alliesenemieschoices(message, page)
    local key = "(#) " .. message
    return {[key]=page, _keys={key}}
end

M.windtunnel = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        local secondarystatistics = character.secondarystatistics
        local buff = 3 * attacks.statistic(character, "power")
        local retaliation = secondarystatistics.retaliation
        retaliation.windtunnel = retaliation.windtunnel 
            and retaliation.windtunnel + buff 
            or buff
        table.insert(
            combatstate.combatlog,
            "Whirls of whistling wind whip around " .. character.combatname .. "!"
        )
        powerutils.inflictcost(combatstate, M.windtunnel)
    end
)

M.windtunnel.cost = 5
M.windtunnel.damage = 0
M.windtunnel.numtargets = 2
M.windtunnel.stats = {"power"}
function M.windtunnel.description(character)
    local retaliation = 3 * attacks.statistic(character, "power")
    return "Gives Tempestas *" .. retaliation .. "* points of retaliation."
end

-- Yomama is a summoning spell that allows Tempestas to use half her current
-- energy to effectively create a copy of herself.
M.yomama = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local character = combatstateops.activecharacter(combatstate)
        table.insert(
            combatstate.combatlog,
            "Hurricane force winds swirl into the shape of a giant eagle!"
        )
        powerutils.summon(combatstate, character, "Yomama", "Y")
    end
)
M.yomama.damage = 0
M.yomama.targets = combatstateops.activeallies
M.yomama.cost = 0
M.yomama.numtargets = 0
M.yomama.stats = {"strength", "speed", "power"}
M.yomama.description = "Tempestas weaves air into an autonomous raptor (the bird not the dinosaur) with all "
    .. " her strength and powers. However, she does have to split her energy between herself and "
    .. " her bird."


-------------------- Twister
--[[
-- Inflicts a debuff on all characters except Tempestas, and a buff
-- on Tempestas.
--
-- On all other characters, twister does tempestas.power/2 damage to
-- them at the end of their turn if they are inflicted with Away. Note
-- that this ignores retaliation, but not shield.
--
-- On Tempestas, at the start of her turn, she is cured of Away, and has
-- her attack modifier increased by away.duration * (tempestas.power/2)
-- for one turn. It also increases Tempestas' by 2.
--]]

M.twister = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local activecharacter = combatstateops.activecharacter(combatstate)
        local duration = attacks.statistic(activecharacter, "power")
        table.insert(
            combatstate.combatlog,
            F"{activecharacter.combatname} swirls a great twister into the air above them!"
        )
        for _, character in ipairs(combatstate.characters) do
            statusops.addstatus(
                character,
                statuses.twister,
                duration,
                activecharacter,
                nil,
                combatstate
            )
        end
        powerutils.inflictcost(combatstate, M.twister)
    end
)

M.twister.damage = 0
M.twister.targets = function(combatstate) return combatstate.characters end
M.twister.cost = 5
M.twister.numtargets = powers.ALL_TARGETS
M.twister.stats = {"power"}
function M.twister.description(character)
    local power = attacks.statistic(character, "power")
    local duration = power
    local damage = math.floor(power / 2)
    return "Tempestas inflicts all targets with twister for "
        .. F"*{duration}* turns. If characters are inflicted with Away, they take "
        .. F"Away's duration x *{damage}* damage at the end of their turn. "
        .. "This ignores defense, and cannot be retaliated against, though "
        .. "shields block it. "
        .. "At the start of her turn, Tempestas is cured of Away and "
        .. "receives an attack bonus equal to "
        .. F"Away's duration x *{damage}* for one turn."
end

--[[
-- Airburst inflicts Away on all targets for power / 10 turns.
--]]
M.airburst = pages.page(
    nil,
    nil,
    nil,
    function(combatstate)
        local activecharacter = combatstateops.activecharacter(combatstate)
        local power = attacks.statistic(activecharacter, "power")
        local duration = 1 + math.floor(power / 10)
        table.insert(
            combatstate.combatlog,
            F"{activecharacter.combatname} unleashes a gust of air, "
                .. " flinging everyone into the air!"
        )
        for _, character in ipairs(combatstate.characters) do
            local flyduration = duration
            if statusops.statusduration(character, statuses.fly) then
                flyduration = flyduration * 2
            end
            statusops.addstatus(
                character,
                statuses.away,
                flyduration, 
                combatstate
            )
        end
    end
)

M.airburst.damage = 0
M.airburst.targets = function(combatstate) return combatstate.characters end
M.airburst.cost = 3
M.airburst.numtargets = powers.ALL_TARGETS
M.airburst.stats = {"power"}
function M.airburst.description(character)
    local power = attacks.statistic(character, "power")
    local duration = 1 + math.floor(power/10)
    return F"Tempestas inflicts away for *{duration}* "
        .. "turns on all targets (including allies)."
end

return M
