--[[
-- Copyright 2016 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- stERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
-- Contains a collection of functions for working with tables in the context
-- of Scarlet.
--]]

local M = {}
local scarletfunctions = require 'scarletfunctions'

--[[
-- Returns a deep copy of the specified table.
--
-- table - The table to copy
-- alreadyseentables - A table that maps tables to their copies. This is used
--  to ensure that references are preserved (so if two keys in the same table
--  are pointing to the same table, they continue pointing to the same table).
--  This is for internal use in the function itself, and should not be provided
--  when invoking it
--
-- returns a deep copy of the specified table
--]]
function M.copy(table, alreadyseentables)
    alreadyseentables = alreadyseentables or {}
    local copy = {}
    for key, value in pairs(table) do
        local newValue
        if type(value) == 'table' then
            newValue = alreadyseentables[value]
                or M.copy(value, alreadyseentables)
            alreadyseentables[value] = newValue
        else
            newValue = value
        end
        copy[key] = newValue
    end
    return copy
end

--[[
-- Returns a shallow copy of the specified table
-- ]]
function M.shallowcopy(table)
    local copy = {}
    for key, value in pairs(table) do
        copy[key] = value
    end
    return copy
end

--[[
-- Returns the number of key-value pairs in the table
--]]
function M.size(table)
    local count = 0
    for key in pairs(table) do
        count = count + 1
    end
    return count
end

--[[
-- Returns the largest number in the given list.
--]]
function M.max(list)
    return math.max(scarletfunctions.unpack(list))
end

--[[
-- Returns the smallest number in the given list
--]]
function M.min(list)
    local min = math.huge
    for _, number in ipairs(list) do
        if number < min then
            min = number
        end
    end
    return min
end

-- [[
-- Returns a function that takes a table of numbers and computes the sum of
-- the numbers in the table.
-- ]]
function M.sum(list)
    local sum = 0
    for _, number in ipairs(list) do
        sum = sum + number
    end
    return sum
end

--[[
-- Tests if two values are deeply equal.
--]]
function M.equal(data1, data2)
    if type(data1) ~= "table" or type(data2) ~= "table" then
        return data1 == data2
    end
    return data1 == data2
        or M.reduce(
            true,
            function (equalSoFar, value, key)
                return equalSoFar and M.equal(value, data2[key])
            end
        )(data1)
end

--[[
-- Given a value, and a table, returns all of the keys that map to that
-- value. Returns nil if table is nil, or the specified value is not mapped
-- to by any _keys.
--]]
function M.reverseLookup(value, map)
    if not map then
        return nil
    end
    local keyList = {}
    for key, value1 in pairs(map) do
        if M.equal(value1, value) then
            table.insert(keyList, key)
        end
    end
    return scarletfunctions.unpack(keyList)
end

--[[
-- Given two lists, concatenates the second list to the
-- end of the first list.
--]]
function M.concat(list1, list2)
    for _, value in ipairs(list2) do
        table.insert(list1, value)
    end
end

--[[
-- Given an iterator, dumps the results into a list, where the first entry
-- in the list is the first value returned by the iterator, and so on.
--
-- Note that this assumes that the iterator only returns a single value.
--]]
function M.dumpiterator(iterator)
    local list = {}
    for value in iterator do
        list[#list+1] = value
    end
    return list
end

local function indentedstring(value, indentlevel)
    local indent = string.rep(" ", indentlevel * 2)
    if type(value) ~= "table" then
        return tostring(value)
    end
    local result = ""
    for k, v  in pairs(value) do
        result = result
            .. indent
            .. "[" .. indentedstring(k, 0) .. "]"
            .. " => "
            .. indentedstring(v, indentlevel + 1) .. "\n"
    end
    return "\n" .. result
end

-- [[
-- A function for building a deep string representation of a given value
-- ]]
function M.string(value)
    return indentedstring(value, 0)
end


--[[
-- Given a list, and a value that may be in the list, returns the index
-- of the value in the list, or nil if the value is not in the list.
-- ]]
function M.find(list, value)
    for i, listvalue in ipairs(list) do
        if listvalue == value then
            return i
        end
    end
    return nil
end

function M.slice(list, first, last, step)
    first = first or 1
    last = last or #table
    step = step or 1
    local sliced = {}
    for i = first, last, step do
        table.insert(sliced, list[i])
    end
    return sliced
end

-- [[
-- Given a list and a value, removes the first instance of the value from the list.
--
-- list - The list to remove the value from
-- value - The value to remove
-- ]]
function M.remove(list, value)
    for i, v in ipairs(list) do
        if v == value then
            table.remove(list, i)
            return
        end
    end
end

return M
